import scipy as s

from .variational_nodes import Variational_Node
from .nodes import Constant_Node

"""
This module defines nodes that are a mix of variational and constant. 
Examples:
- For some factors the spike and slab sparsity parameter (theta) is fixed, whereas for others is learnt.
- For some factors the weights can be made sparse whereas for others we might not want to do any sparsity (i.e. mean or covariates)
"""

class Mixed_Theta_Nodes(Variational_Node, Constant_Node):
    """
    Class for a mixture of LearningTheta and ConstantTheta nodes.
    For a total of K factors, some (Klearn) will learn Theta whereas for the others (Kconst) it will be constant
        K = Klearn + Kconst
    """
    def __init__(self, LearnTheta, ConstTheta, idx):
        # Inputs:
        # - LearnTheta: Theta_Node with dimensions (Klearn,)
        # - ConstTheta: Theta_Constant_Node with dimensions (D,Kconst) or (Kconst,1) - NOT IMPLEMENTED YET -
        # - idx: list or numpy array indicating which factors are LearnTheta(idx=1. or idx=True) and which are ConstTheta(idx=0. or idx=False)
        self.constTheta = ConstTheta
        self.learnTheta = LearnTheta

        self.K = ConstTheta.dim[0] + LearnTheta.dim[0]
        self.D = ConstTheta.dim[0]

        self.idx = idx
        
    def precompute(self):
        self.constTheta.precompute()
        self.learnTheta.precompute()

    def addMarkovBlanket(self, **kargs):
        # SHOULD WE ALSO ADD MARKOV BLANKET FOR CONSTHTETA???
        self.learnTheta.addMarkovBlanket(**kargs)

    def getExpectations(self):
        
        # Get expectations from ConstTheta nodes (D,Kconst)
        Econst = self.constTheta.getExpectations()

        # Get expectations from LearnTheta nodes
        Elearn = self.learnTheta.getExpectations()

        # Concatenate
        # Concatenate expectations to (D,K)
        E = s.concatenate((Econst["E"], Elearn["E"]), axis=0)
        lnE = s.concatenate((Econst["lnE"], Elearn["lnE"]), axis=0)
        lnEInv = s.concatenate((Econst["lnEInv"], Elearn["lnEInv"]), axis=0)        

        # Permute to the right order given by self.idx
        # idx = s.concatenate((s.nonzero(1-self.idx)[0],s.where(self.idx)[0]), axis=0)
        # E, lnE, lnEinv = E[idx], lnE[idx], lnEInv[idx]

        return dict({'E':E, 'lnE':lnE, 'lnEInv':lnEInv})

    def getExpectation(self):
        return self.getExpectations()['E']

    def updateExpectations(self):
        self.learnTheta.updateExpectations()

    def updateParameters(self):
        # the argument contains the indices of the non_annotated factors
        self.learnTheta.updateParameters(s.nonzero(self.idx)[0])

    def calculateELBO(self):
        return self.learnTheta.calculateELBO()

    def removeFactors(self, *idx):
        for i in idx:
            if self.idx[idx] == 1:
                self.learnTheta.removeFactors(s.where(i == s.nonzero(self.idx)[0])[0])
            else:
                self.constTheta.removeFactors(s.where(i == s.nonzero(1-self.idx)[0])[0])
            self.idx = self.idx[s.arange(self.K)!=i]
            self.K -= 1




"""
Module to define multi-view nodes. 
A multi-view node is simply a node that is defined for several views. For example, the weights W or the data Y, but not the latent variables Z.

Types of multi-view nodes:
- Multiview_Variational_Node: variational nodes present in all views
- Multiview_Constant_Node: multiview nodes that are just constant values
- Multiview_Mixed_Node: a node which is variational for some views (i.e. gaussian views) and constant for others (i.e. non-gaussian views)

All multiview nodes have the following main attributes:
- M: total number of views
- activeM: in some occasions a particular node is active in only a subset of views. For example, we could activate spike-and-slab in one view but not in the other.
- nodes: a list with the (single-view) nodes
"""

import scipy as s

from .nodes import Node
from .variational_nodes import Variational_Node


class Multiview_Node(Node):
    """General class for a multiview node"""
    def __init__(self, M, *nodes):
        """
        PARAMETERS
        ----------
        M: int
            total number of views
        nodes: list
            list of M nodes, which must be instances or children of the 'Node' class. If the node is not defined in view m, then index m is set to None.
        """
        self.M = M
        self.activeM = [ m for m,node in enumerate(nodes) if node is not None]
        self.nodes = nodes

    def addMarkovBlanket(self, **kwargs):
        """Method to define the Markov blanket"""
        # assert len(kwargs.values()) == len(self.activeM), "The markov blanket of a multiview node should be a dictionary where the key is the name of the node and the value is a list of nodes of length M"
        for k,v in kwargs.items(): 
            for m in self.activeM: 
                if hasattr(self.nodes[m], 'markov_blanket'):
                    if k in self.nodes[m].markov_blanket.keys():
                        print("Error: " + str(k) + " is already in the markov blanket of " + str(self.nodes[m]))
                    else:
                        if isinstance(v,Multiview_Node):
                            self.nodes[m].markov_blanket[k] = v.getNodes()[m]
                        else:
                            self.nodes[m].markov_blanket[k] = v
                else:
                    self.nodes[m].addMarkovBlanket( **{ k: (v.getNodes()[m] if isinstance(v,Multiview_Node) else v) } )

    def getMarkovBlanket(self):
        print("Error: Multiview nodes do not have a markov blanket, use the single-view nodes")
        exit()
        
    def removeFactors(self,idx):
        """Method to remove factors from the node

        PARAMETERS
        ----------
        idx: ndarray
            indices of the factors to be removed
        """
        for m in self.activeM: self.nodes[m].removeFactors(idx)

    def getNodes(self):
        """Method to get the nodes"""
        return self.nodes
        
    def getMask(self):
        """Method to get the masks"""
        return [ self.nodes[m].getMask() for m in self.activeM ]

    def getExpectation(self):
        """Method to get the first moments (expectation)"""
        return [ self.nodes[m].getExpectation() for m in self.activeM ]

    def getExpectations(self):
        """Method to get all moments"""
        return [ self.nodes[m].getExpectations() for m in self.activeM ]

    def getParameters(self):
        """Method to get  the parameters"""
        return [ self.nodes[m].getParameters() for m in self.activeM ]

    def updateDim(self, axis, new_dim, m=None):
        """Method to update the dimensionality of the node

        PARAMETERS
        ----------
        axis: int
        new_dim: int
        m: iterable
            views to update
        """
        assert s.all(m in self.activeM), "Trying to update the dimensionality of a node that doesnt exist in a view"
        M = self.activeM if m is None else m
        for m in M: self.nodes[m].updateDim(axis,new_dim)

    def precompute(self):
        for m in self.activeM:
            self.nodes[m].precompute()

class Multiview_Variational_Node(Multiview_Node, Variational_Node):
    """General class for multiview variational nodes."""
    def __init__(self, M, *nodes):
        Multiview_Node.__init__(self, M, *nodes)
        for node in nodes: assert isinstance(node, Variational_Node)

    def update(self):
        """ Method to update both parameters and expectations of the node"""
        for m in self.activeM:
            self.nodes[m].updateParameters()
            self.nodes[m].updateExpectations()
    def updateExpectations(self):
        """Method to update expectations using current estimates of the parameters"""
        for m in self.activeM: self.nodes[m].updateExpectations()
    def updateParameters(self):
        """Method to update parameters using current estimates of the expectations"""
        for m in self.activeM: self.nodes[m].updateParameters()
    def calculateELBO(self):
        """Method to calculate variational evidence lower bound"""
        lb = [ self.nodes[m].calculateELBO() for m in self.activeM ]
        return sum(lb)

class Multiview_Constant_Node(Multiview_Node):
    """General class for multiview local nodes"""
    def __init__(self, M, *nodes):
        Multiview_Node.__init__(self, M, *nodes)

    def getValues(self):
        """Method to retun the values of the node"""
        return [ self.nodes[m].getValue() for m in self.activeM ]

class Multiview_Mixed_Node(Multiview_Constant_Node, Multiview_Variational_Node):
    """General Class for multiview nodes that contain both variational and constant nodes"""
    def __init__(self, M, *nodes):
        # M: number of views
        # nodes: list of M 'Node' instances
        Multiview_Node.__init__(self, M, *nodes)


    def update(self):
        """Method to update values of the nodes"""
        for m in self.activeM: 
            self.nodes[m].update()

    def calculateELBO(self):
        """Method to calculate variational evidence lower bound
        The lower bound of a multiview node is the sum of the lower bound of its corresponding single view variational nodes
        """
        lb = 0
        for m in self.activeM:
            if isinstance(self.nodes[m],Variational_Node):
                lb += self.nodes[m].calculateELBO()
        return lb

